from typing import Generic, Iterable, TypeAlias, TypeVar

from rainbowadn.collection.comparison import *
from rainbowadn.core import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._flowstandard import *

__all__ = ('FlowBlock', 'FlowBlockFactory', 'FlowBlockVerification',)

LinkT = TypeVar('LinkT')

FBL: TypeAlias = 'FlowBlock[LinkT]'
Index: TypeAlias = FlowStandard[FBL]


class FlowBlock(Generic[LinkT], RecursiveMentionable):
    def points(self) -> Iterable[HashPoint]:
        return [*self.previous.points(), *self.index.points(), self.link]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.index) + bytes(self.link)

    def __factory__(self) -> RainbowFactory[FBL]:
        return FlowBlockFactory(self.link.factory)

    def __init__(
            self,
            previous: NullableReference[FBL],
            index: Index,
            link: HashPoint[LinkT],
    ):
        assert isinstance(previous, NullableReference)
        assert isinstance(index, FlowStandard)
        self.previous = previous
        self.index = index
        self.link = link

    async def outer(self) -> Index:
        return FlowStandard(
            (
                await self.index.protocolized.tree.add(
                    HashPoint.of(self)
                )
            ).protocolized()
        )

    async def verify_outer_matches(self, index: Index) -> bool:
        assert_eq(await self.outer(), index)
        return True

    @classmethod
    async def outer_of(cls, factory: RainbowFactory[LinkT], reference: NullableReference[FBL]) -> Index:
        if reference.null():
            return FlowStandardFactory.empty(FlowBlockFactory(factory), HashComparator(Fail()))
        else:
            return await (await reference.resolve()).outer()

    @classmethod
    async def link_of(cls, reference: NullableReference[FBL]) -> Nullable[LinkT]:
        if reference.null():
            return Null()
        else:
            return NotNull(await (await reference.resolve()).link.resolve())

    async def add(self, link: HashPoint[LinkT]) -> FBL:
        return FlowBlock(
            NullableReference.off(self),
            await self.outer(),
            link
        )

    @classmethod
    async def add_to(cls, reference: NullableReference[FBL], link: HashPoint[LinkT]) -> FBL:
        return FlowBlock(
            reference,
            await cls.outer_of(link.factory, reference),
            link
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        previous_str: str
        link_str: str
        previous_str, link_str = await gather(
            self.previous.str(tab),
            hash_point_format(self.link, tab)
        )
        assert isinstance(previous_str, str)
        assert isinstance(link_str, str)
        return f'{previous_str}' \
               f'{tabulate(tab)}(index)' \
               f'{tabulate(tab)}{link_str}'


class FlowBlockFactory(RainbowFactory[FBL], Generic[LinkT]):
    def __init__(self, factory: RainbowFactory[LinkT]):
        assert isinstance(factory, RainbowFactory)
        self.factory = factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> FBL:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return FlowBlock(
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            FlowStandardFactory.of(self, HashComparator(Fail())).from_bytes(
                source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver
            ),
            ResolverOrigin(self.factory, source[2 * HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    def loose(self) -> RainbowFactory[FBL]:
        return self


class FlowBlockIndexedVerification(
    Verification[HashPoint[FBL]],
    Generic[LinkT]
):
    def __init__(
            self,
            index: Index,
            verification: Verification[tuple[Nullable[HashPoint[LinkT]], HashPoint[LinkT]]],
    ):
        assert isinstance(index, FlowStandard)
        assert isinstance(verification, Verification)
        self.index = index
        self.verification = verification

    async def _verify_empty_block(self, block: FBL) -> bool:
        assert isinstance(block, FlowBlock)
        assert_trues(
            await gather(
                self.verification.verify(
                    (Null(), block.link)
                ),
                block.index.verify_empty(),
            )
        )
        return True

    async def _verify_previous_block(self, previous: FBL, block: FBL) -> bool:
        assert isinstance(previous, FlowBlock)
        assert isinstance(block, FlowBlock)
        assert_trues(
            await gather(
                self.verification.verify(
                    (NotNull(previous.link), block.link)
                ),
                previous.verify_outer_matches(block.index),
            )
        )
        return True

    async def _verify_previous(self, previous: HashPoint[FBL], block: FBL) -> bool:
        assert_true(await self._verify_previous_block(await previous.resolve(), block))
        return True

    async def _verify(self, block: FBL) -> bool:
        assert isinstance(block, FlowBlock)
        if block.previous.null():
            assert_true(await self._verify_empty_block(block))
        else:
            assert_true(await self._verify_previous(block.previous.hashpoint(), block))
        return True

    async def verify(self, element: HashPoint[FBL]) -> bool:
        assert isinstance(element, HashPoint)
        assert_true(await self._verify(await element.resolve()))
        return True

    def loose(self) -> Verification[HashPoint[FBL]]:
        return self


class FlowBlockVerification(
    Verification[Index],
    Generic[LinkT]
):
    def __init__(
            self,
            verification: Verification[tuple[Nullable[HashPoint[LinkT]], HashPoint[LinkT]]],
    ):
        assert isinstance(verification, Verification)
        self.verification = verification

    async def verify(self, element: Index) -> bool:
        assert_true(
            await ReduceVerification(
                FlowBlockIndexedVerification(element, self.verification).loose()
            ).verify(
                await element.reducer()
            )
        )
        return True

    def loose(self) -> Verification[Index]:
        return self
