import heapq
from io import BytesIO
from typing import Callable, Generic, Iterable, Optional, Type, TypeAlias, TypeVar

from rainbowadn.core import *
from .inlining import *

__all__ = ('IAuto', 'IAutoFactory',)

_IList: TypeAlias = list[tuple[int, Mentionable]]
_UList: TypeAlias = list[tuple[int, HashPoint]]
_SList: TypeAlias = list[tuple[int, RainbowFactory, int]]
_VList: TypeAlias = list[tuple[int, RainbowFactory]]
_MCall: TypeAlias = Callable[[bytes, HashResolver, _IList, _UList], None]
_MTuple: TypeAlias = tuple[int, _MCall, int]
_IAuto = TypeVar('_IAuto')


class IAuto(RecursiveMentionable):
    def __init__(
            self,
            inlined: _IList,
            uninlined: _UList
    ):
        assert isinstance(inlined, list)
        assert isinstance(uninlined, list)
        for index, mentionable in inlined:
            assert isinstance(index, int)
            assert isinstance(mentionable, Mentionable)
        for index, hashpoint in uninlined:
            assert isinstance(index, int)
            assert isinstance(hashpoint, HashPoint)
        self.inlined = inlined
        self.uninlined = uninlined

    def points(self) -> Iterable[HashPoint]:
        inlined_points: list[tuple[int, HashPoint]] = []
        uninlined_points: list[tuple[int, HashPoint]] = []
        for index, inlined in self.inlined:
            if isinstance(inlined, RecursiveMentionable):
                for point in inlined.points():
                    assert isinstance(point, HashPoint)
                    inlined_points.append((index, point))
        for index, uninlined in self.uninlined:
            uninlined_points.append((index, uninlined))
        merged_points: Iterable[tuple[int, HashPoint]] = heapq.merge(inlined_points, uninlined_points)
        for _, point in merged_points:
            assert isinstance(point, HashPoint)
            yield point

    def __bytes__(self):
        inlined_bytes: list[tuple[int, bytes]] = []
        uninlined_bytes: list[tuple[int, bytes]] = []
        for index, inlined in self.inlined:
            inlined_bytes.append((index, bytes(inlined)))
        for index, uninlined in self.uninlined:
            uninlined_bytes.append((index, bytes(uninlined)))
        merged_bytes: Iterable[tuple[int, bytes]] = heapq.merge(inlined_bytes, uninlined_bytes)
        return b''.join(source for _, source in merged_bytes)

    def __factory__(self: _IAuto) -> RainbowFactory[_IAuto]:
        sized: _SList = []
        inlined_unsized: _VList = []
        uninlined_unsized: _VList = []
        for index, inlined in self.inlined:
            factory = inlined.__factory__()
            size = Inlining.factory_size(factory)
            if size is None:
                inlined_unsized.append((index, factory))
            else:
                assert isinstance(size, int)
                sized.append((index, factory, size))
        for index, uninlined in self.uninlined:
            uninlined_unsized.append((index, uninlined.auto_f))
        merged_unsized: Iterable[tuple[int, RainbowFactory]] = heapq.merge(inlined_unsized, uninlined_unsized)
        return IAutoFactory(
            type(self),
            sized,
            list(merged_unsized)
        )

    def hashpoints(self) -> Iterable[HashPoint]:
        inlined_hashpoints: list[tuple[int, HashPoint]] = []
        uninlined_hashpoints: list[tuple[int, HashPoint]] = []
        for index, inlined in self.inlined:
            inlined_hashpoints.append((index, HashPoint.of(inlined)))
        for index, uninlined in self.uninlined:
            uninlined_hashpoints.append((index, uninlined))
        merged_hashpoints: Iterable[tuple[int, HashPoint]] = heapq.merge(inlined_hashpoints, uninlined_hashpoints)
        for _, hashpoint in merged_hashpoints:
            yield hashpoint

    async def values(self) -> Iterable[Mentionable]:
        return await gather(*map(HashPoint.resolve, self.hashpoints()))

    @classmethod
    def auto_f(cls: Type[_IAuto], *factories: RainbowFactory) -> RainbowFactory[_IAuto]:
        assert issubclass(cls, IAuto)
        return IAutoFactory.of(cls, *factories)

    @classmethod
    def auto_of(cls: Type[_IAuto], *values: Mentionable) -> _IAuto:
        assert issubclass(cls, IAuto)
        sized: _IList = []
        unsized: _IList = []
        for index, value in enumerate(values):
            assert isinstance(value, Mentionable)
            factory = value.__factory__()
            if Inlining.factory_size(factory) is None:
                unsized.append((index, value))
            else:
                sized.append((index, value))
        inlined: _IList
        uninlined: _UList
        match unsized:
            case [(_, _)]:
                inlined = list(heapq.merge(sized, unsized))
                uninlined = []
            case _:
                inlined = sized
                uninlined = [(index, HashPoint.of(value)) for index, value in unsized]
        return cls(inlined, uninlined)

    @classmethod
    async def _utuple_to_ituple(cls, utuple: tuple[int, HashPoint]) -> tuple[int, Mentionable]:
        assert isinstance(utuple, tuple)
        index, hashpoint = utuple
        return index, await hashpoint.resolve()

    @classmethod
    async def auto_off(cls: Type[_IAuto], *hashpoints: HashPoint) -> _IAuto:
        assert issubclass(cls, IAuto)
        sized: _UList = []
        unsized: _UList = []
        for index, hashpoint in enumerate(hashpoints):
            assert isinstance(hashpoint, HashPoint)
            factory = hashpoint.factory
            if Inlining.factory_size(factory) is None:
                unsized.append((index, hashpoint))
            else:
                sized.append((index, hashpoint))
        inlined: _IList
        uninlined: _UList
        match unsized:
            case [(_, _)]:
                inlined = list(await gather(*map(cls._utuple_to_ituple, heapq.merge(sized, unsized))))
                uninlined = []
            case _:
                inlined = list(await gather(*map(cls._utuple_to_ituple, sized)))
                uninlined = unsized
        return cls(inlined, uninlined)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        formatted = await gather(*(hash_point_format(hashpoint, tab) for hashpoint in self.hashpoints()))
        return f'{tabulate(tab)}'.join(formatted)


class IAutoFactory(Inlining[IAuto], Generic[_IAuto]):
    def __init__(
            self,
            typed: Type[_IAuto],
            sized: _SList,
            unsized: _VList
    ):
        assert issubclass(typed, IAuto)
        assert isinstance(sized, list)
        assert isinstance(unsized, list)
        self.typed = typed
        for index, factory, size in sized:
            assert isinstance(index, int)
            assert isinstance(factory, RainbowFactory)
            assert isinstance(size, int)
        for index, factory in unsized:
            assert isinstance(index, int)
            assert isinstance(factory, RainbowFactory)
        self.sized = sized
        self.unsized = unsized
        match unsized:
            case [(infix_index, infix_factory)]:
                self.infix = True
                self.infix_index: int = infix_index
                self.infix_factory: RainbowFactory = infix_factory
                self.prefix: _SList = list(
                    (index, factory, size) for index, factory, size in sized if index < infix_index
                )
                self.prefix_size = self._affix_size(self.prefix)
                assert isinstance(self.prefix_size, int)
                self.postfix: _SList = list(
                    (index, factory, size) for index, factory, size in sized if index > infix_index
                )
                self.postfix_size = self._affix_size(self.postfix)
                assert isinstance(self.postfix_size, int)
            case _:
                self.infix = False
                self.merged = self._merged()
                assert isinstance(self.merged, list)

    @classmethod
    def of(cls, typed: Type[_IAuto], *factories: RainbowFactory) -> Inlining[_IAuto]:
        sized: _SList = []
        unsized: _VList = []
        for index, factory in enumerate(factories):
            assert isinstance(factory, RainbowFactory)
            size = Inlining.factory_size(factory)
            if size is None:
                unsized.append((index, factory))
            else:
                assert isinstance(size, int)
                sized.append((index, factory, size))
        return cls(typed, sized, unsized)

    @classmethod
    def _affix_size(cls, affix: _SList) -> int:
        assert isinstance(affix, list)
        return sum(size for _, _, size in affix)

    @classmethod
    def _unsized_mtuple(cls, index: int, factory: RainbowFactory) -> _MTuple:
        assert isinstance(index, int)
        assert isinstance(factory, RainbowFactory)

        def wrap(source: bytes, resolver: HashResolver, inlined: _IList, uninlined: _UList) -> None:
            assert isinstance(source, bytes)
            assert isinstance(resolver, HashResolver)
            assert isinstance(inlined, list)
            assert isinstance(uninlined, list)
            assert_eq(len(source), HashPoint.HASH_LENGTH)
            uninlined.append((index, ResolverMetaOrigin(resolver).hash_point(factory, source)))

        return index, wrap, HashPoint.HASH_LENGTH

    @classmethod
    def _sized_mtuple(cls, index: int, factory: RainbowFactory, size: int) -> _MTuple:
        assert isinstance(index, int)
        assert isinstance(factory, RainbowFactory)
        assert isinstance(size, int)

        def wrap(source: bytes, resolver: HashResolver, inlined: _IList, uninlined: _UList) -> None:
            assert isinstance(source, bytes)
            assert isinstance(resolver, HashResolver)
            assert isinstance(inlined, list)
            assert isinstance(uninlined, list)
            assert_eq(len(source), size)
            inlined.append((index, factory.from_bytes(source, resolver)))

        return index, wrap, size

    def _merged(self) -> list[tuple[_MCall, int]]:
        unsized_mergeable: list[_MTuple] = []
        sized_mergeable: list[_MTuple] = []
        for index, factory in self.unsized:
            unsized_mergeable.append(
                self._unsized_mtuple(index, factory)
            )
        for index, factory, size in self.sized:
            sized_mergeable.append(
                self._sized_mtuple(index, factory, size)
            )
        merged: Iterable[_MTuple] = heapq.merge(unsized_mergeable, sized_mergeable)
        return [(method, size) for _, method, size in merged]

    def size(self) -> Optional[int]:
        if self.infix:
            return None
        else:
            return sum(size for _, size in self.merged)

    @classmethod
    def _parse_affix(
            cls, source: bytes, resolver: HashResolver, affix: _SList
    ) -> Iterable[tuple[int, Mentionable]]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        assert isinstance(affix, list)
        reader = BytesIO(source)
        for index, factory, size in affix:
            yield index, factory.from_bytes(reader.read(size), resolver)
        assert not reader.read()

    def _parse_infix(self, source: bytes, resolver: HashResolver) -> Iterable[tuple[int, Mentionable]]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        yield from self._parse_affix(source[:self.prefix_size], resolver, self.prefix)
        postfix_start = len(source) - self.postfix_size
        yield self.infix_index, self.infix_factory.from_bytes(source[self.prefix_size:postfix_start], resolver)
        yield from self._parse_affix(source[postfix_start:], resolver, self.postfix)

    def _parse_merged(self, source: bytes, resolver: HashResolver) -> _IAuto:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        reader = BytesIO(source)
        inlined: _IList = []
        uninlined: _UList = []
        for method, size in self.merged:
            method(reader.read(size), resolver, inlined, uninlined)
        assert not reader.read()
        return self.typed(inlined, uninlined)

    def from_bytes(self, source: bytes, resolver: HashResolver) -> _IAuto:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        if self.infix:
            return self.typed(list(self._parse_infix(source, resolver)), [])
        else:
            return self._parse_merged(source, resolver)
